package com.fundi.framework.common.cslcomms;

/*
 *  Common Services Library - Framework plug-in.
 *  (C) Copyright Fundi Software 2011, 2016.  Fundi Confidential.
 *  All rights reserved.  Use, duplication or disclosure restricted. 
 */

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Iterator;

import com.fundi.framework.common.nl1.Messages;
import com.fundi.framework.common.util.StringEditor;
import com.fundi.framework.common.util.StringReplacer;

public class CSLField implements Serializable {
	
	//Eye catchers.
	public static final String EC_COPYRIGHT = "(C) Copyright 2011, 2016 Fundi Software.  Fundi Confidential."; //$NON-NLS-1$
	public static final String EC_APAR = "3.1.1.00_00014"; //$NON-NLS-1$
	public static final String EC_FMID = "H30SA10"; //$NON-NLS-1$
	public static final String EC_PID = "5655-CSL"; //$NON-NLS-1$
	
	private static final long serialVersionUID = 1L;
	
	private CSLFieldDefn defn = null;
	private String value;

	private static StringReplacer	unescapeReplacer = new StringReplacer(new String[] {
		Messages.getString("CSLField_5"), 
		Messages.getString("CSLField_7"), 
		Messages.getString("CSLField_9"), 
		Messages.getString("CSLField_11"),
		Messages.getString("CSLField_13"),
		Messages.getString("CSLField_15"),
		Messages.getString("CSLField_17"),
		Messages.getString("CSLField_19"),
		Messages.getString("CSLField_21")
	}, new String[] {
		Messages.getString("CSLField_6"),
		Messages.getString("CSLField_8"),
		Messages.getString("CSLField_10"),
		Messages.getString("CSLField_12"),
		Messages.getString("CSLField_14"),
		Messages.getString("CSLField_16"),
		Messages.getString("CSLField_18"),
		Messages.getString("CSLField_20"),
		Messages.getString("CSLField_22")
	});
	
	public CSLField(ArrayList<CSLFieldDefn>defns, String rspFld) {
		
		super();
		
		//This constructor is for type-2 command fields.
		Iterator<CSLFieldDefn> it;
		String label;
		int posLparen;
		
		posLparen = rspFld.indexOf(Messages.getString("CSLField_4")); //$NON-NLS-1$
		label = rspFld.substring(0, posLparen);
		
		//A field definition supplied from GPL could contain escaped characters.
		value = unescape(rspFld.substring(posLparen + 1, rspFld.length() - 1));
		if (value.length() > 0) {
			value = (new StringEditor(value)).rtrim();
		}
		
		//The rspFld is a component of the <rsp> object within the <cmdrspdata> object.
		//It has the format fldName(fldValue);
		for (it = defns.iterator(); it.hasNext();) {
			defn = it.next();
			if (defn.getSlbl().equals(label)) {
				break;
			}
		}
	}
	
	public CSLField(ArrayList<CSLFieldDefn>defns, String name, String value) {
		
		super();
		
		//This constructor is for type-1 command fields.
		Iterator<CSLFieldDefn> it;
		
		this.value = value;
		
		for (it = defns.iterator(); it.hasNext();) {
			defn = it.next();
			if (defn.getSlbl().equals(name)) {
				break;
			}
		}
	}

	private String unescape(String origValue) {
		return unescapeReplacer.replace(origValue);
	}
	
	public String getValue() {
		
		return value;
	}
	
	public int getIntValue() {
		
		return Integer.parseInt(value);
	}
	
	public String getSlbl() {
		
		return defn.getSlbl();
	}
	
	public String getLlbl() {
		 
		return defn.getLlbl();
	}

	public void setValue(String value) {
		
		this.value = value;
	}
	
	public boolean isKey() {
		
		return defn.isKey();
	}
	
	public String getDataType() {
		
		return defn.getDtype();
	}
	
	public String getAlignment() {
		
		return defn.getAlign();
	}
	
	public int getMaxLength() {
		
		return defn.getMaxLength();
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((defn == null) ? 0 : defn.hashCode());
		result = prime * result + ((value == null) ? 0 : value.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		CSLField other = (CSLField) obj;
		if (defn == null) {
			if (other.defn != null)
				return false;
		} else if (!defn.equals(other.defn))
			return false;
		if (value == null) {
			if (other.value != null)
				return false;
		} else if (!value.equals(other.value))
			return false;
		return true;
	}

	public String getKeyLabel() {
		return defn.getKeyLabel();
	}

	public void setKeyLabel(String keyLabel) {
		defn.setKeyLabel(keyLabel);
	}
}
