package com.fundi.framework.common.cslcomms;

/*
 *  Common Services Library - Framework plug-in.
 *  (C) Copyright Fundi Software 2011, 2016.  Fundi Confidential.
 *  All rights reserved.  Use, duplication or disclosure restricted. 
 */

import java.io.DataInputStream;
import java.io.FilterInputStream;
import java.io.IOException;

public class CSLInputStream extends FilterInputStream {
	
	//Eye catchers.
	public static final String EC_COPYRIGHT = "(C) Copyright 2011, 2016 Fundi Software.  Fundi Confidential."; //$NON-NLS-1$
	public static final String EC_APAR = "3.1.1.00_00014"; //$NON-NLS-1$
	public static final String EC_FMID = "H30SA10"; //$NON-NLS-1$
	public static final String EC_PID = "5655-CSL"; //$NON-NLS-1$
	
	int bytesRead; 
	int segmentLength;
	byte[] prefixBytes = null;
	boolean eofRead = false;
	byte[] segmentBody = null;
	int returnCode = 0;
	int reasonCode = 0;
	DataInputStream in;
	int segmentCount = 0;

	public CSLInputStream(DataInputStream in) {

		super(in);
		
		this.in = in;
		
		eofRead = false;

		readSegment();
	}
	
	private void readSegment() {
		
		try {
			bytesRead = 0;
			segmentLength = 0;
			
			if (getReturnCode() == 4) {
				eofRead = true;
			}
			else {
				segmentLength = in.readInt();
				if (segmentLength == 0) {
					eofRead = true;
				}
				else {
					segmentLength = segmentLength - 4;
				}
			}
			segmentCount++;
		}
		catch (IOException e) {
			eofRead = true;
		}
	}
	
	public int read() throws IOException {
		
		int b;
		
		b = in.read();
		
		//Don't care about anything in the message after the end of the last segment - logically EOF.
		if (b < 0) {
			return b;
		}
		else {
			if (bytesRead < segmentLength) {
				bytesRead++;
				//System.out.println((char) b);
				return b;
			}
			else {
				if (eofRead) {
					return -1;
				}
				else {
					b = in.read();
					bytesRead++;
					//System.out.println((char) b);
					return b;
				}
			}
		}
	}
	
	public int read(byte[] data, int offset, int length) throws IOException {
		
		int bytesToRead;
		int read;

		if ((bytesRead + length) <= segmentLength) {
			bytesToRead = length; 
		}
		else {
			bytesToRead = segmentLength - bytesRead;
			if (bytesToRead <= 0) {
				return -1;
			}
		}
				
		read = in.read(data, offset, bytesToRead);
		if (read < 0) {
			//System.out.println(new String(segmentBody).substring(0, read));
			return read;
		}
		bytesRead = bytesRead + read;
		segmentBody = data;
		
		//System.out.println(new String(segmentBody).substring(0, read));
		return read;
	}
	
	public int read(byte[] data) throws IOException {
		  
		return read(data, 0, data.length);
	}

	public int getReturnCode() {
		return returnCode;
	}

	public int getReasonCode() {
		return reasonCode;
	}
}
