package com.fundi.framework.common.cslcomms;

/*
 *  Common Services Library - Framework plug-in.
 *  (C) Copyright Fundi Software 2011, 2016.  Fundi Confidential.
 *  All rights reserved.  Use, duplication or disclosure restricted. 
 */

import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.net.Socket;
import java.net.SocketException;
import java.util.ArrayList;

import com.fundi.framework.common.instance.AppInstance;
import com.fundi.framework.common.instance.ConnectionProfile;
import com.fundi.framework.common.instance.ConnectionServer;
import com.fundi.framework.common.instance.ConsoleMessage;
import com.fundi.framework.common.nl1.Messages;
import com.fundi.framework.common.util.StringEditor;

public class ConnectionServerDriver {

	//Eye catchers.
	public static final String EC_COPYRIGHT = "(C) Copyright 2011, 2016 Fundi Software.  Fundi Confidential."; //$NON-NLS-1$
	public static final String EC_APAR = "3.1.1.00_00014"; //$NON-NLS-1$
	public static final String EC_FMID = "H30SA10"; //$NON-NLS-1$
	public static final String EC_PID = "5655-CSL"; //$NON-NLS-1$
	
	private AppInstance instance = null;
	private ConnectionServer srvr = null;
	//private Socket socket = null;
	
	private ArrayList<Socket> sockets = new ArrayList<Socket>();
	private ArrayList<String> socketIds = new ArrayList<String>();
	
	private static int xhdrLength = 256;
	
	//Following fields describe the XHDR message format sent to ConnectionServer.
	//See http://intranet/wiki/display/FUD/Request+structure+cheat+sheet for the specification followed here.
	//The payload in the request immediately follows these fields.
	
	private int xhdrTotalMsgLength;  //Total length of message, including payload.
	//XHDR control fields.
	private String xhdrCtlFldEyec;  //Always use XHDR (if you use ASCII, it will perform translation of XHDR)
	private String xhdrCtlFldArclev; //Architecture level: use 0
	private String xhdrCtlFldType; //Always use TCP
	//Fields set by server for TCP/IP client (do not set or clear, just skip 24 Bytes)
	private int xhdrTcpTMTE;  //TCP Message Thread Element
	private byte[] xhdrTcpSKD;  //Local socket descriptor
	private short xhdrTcpPort; //Remote port number
	private String xhdrTcpAddr;  //Remote INET address
	//Session data the client needs to set
	private int xhdrSdLen;  //Estimated length of message (optional, helps the server allocate memory)
	private String xhdrSdSegmctl;  //F : First;  M : Middle  (not first or last); L : Last
	private String xhdrSdTr;  //Y : Yes, payload requires translation (Y if using ASCII); or N for no translation
	//Fields the client needs to set
	private String xhdrCFunc;  //Always use 1 - Product request
	private String xhdrCProduct;  //E.g. FUD or GPL (for ping and delay)
	private short xhdrCCCSID; //819 = x'333'.  //55;  //Use x'37'
	private int xhdrCBufsz;  //Buffer size use 0 to get each response in full, i.e. unsegmented.
	private String xhdrCChkid;  //Can set to zeros (it is echoed back so you can use it for consistency checking)
	private String xhdrCUid;  //User ID (use the current CEX encryption for the user ID)
	private String xhdrCPwd;  //Password (use the current CEX encryption for the password)
	private long xhdrCReserved1;  //For future use
	//Server-maintained fields client must keep and send back on requests (28 bytes)
	private byte[] xhdrSConvTkn;  //CONV Token
	private byte[] xhdrRRSESTkn;  //RSES Token
	private byte[] xhdrSPRDE;  //PRDE entry address
	private byte[] xhdrSConv;  //CONV address
	private byte[] xhdrSRSES;  //RSES entry address
	//Server status information (view only, for diag use information in next section, not this)
	private int xhdrRc;
	private int xhdrRsn;
	private int xhdrInfo;
	private int xhdrInfo2;
	private int xhdrReserved2;
	//Server feedback in text (40 bytes)
	private String xhdrCRc;  //Return code
	private String xhdrCRsn;  //Word 1 = <ModuleID><ReasonCode>
	private String xhdrCInfo; //INFO
	private String xhdrCInfo1;  //INFO
	private long xhdrCReserved3;
	//Server request-response cycle statistics fields
	private String xhdrStaTime;  //Timestamp of XHDR read into server
	private String xhdrStaLSN;  //with server log sequence number
	private String xhdrStoTime;  //Timestamp of XHDR send from server
	private String xhdrStoLSN;  //with server log sequence number
	
	private Xheader xhdr = null;
	private String appResponse = null;
	
	public static final String returnTypeCSLResponse = Messages.getString("ConnectionServerDriver_4"); //$NON-NLS-1$
	public static final String returnTypeCSLObjects = Messages.getString("ConnectionServerDriver_5"); //$NON-NLS-1$
	
	public ConnectionServerDriver(AppInstance instance, ConnectionServer srvr) {
		
		super();
		
		this.instance = instance;
		this.srvr = srvr;
	}
	
	public String getConfigurationInfo() {
		
		StringBuffer info = new StringBuffer(""); //$NON-NLS-1$
		CSLResponse cslResp;
		String cmd = "FUDQUERY PRODUCT CODE(*)"; //$NON-NLS-1$
		int idx;
		ArrayList<CSLObject> objects;
		
		cslResp = new CSLResponse(runCommand(cmd, "FUD"), cmd); //$NON-NLS-1$
		objects = cslResp.getAllObjects();
		
		info.append(Messages.getString("ConnectionServerDriver_62") + srvr.getName() + "\r\n"); //$NON-NLS-1$ //$NON-NLS-2$
		info.append(Messages.getString("ConnectionServerDriver_59") + srvr.getHost() + "\r\n"); //$NON-NLS-1$ //$NON-NLS-2$
		info.append(Messages.getString("ConnectionServerDriver_60") + srvr.getPort() + "\r\n\r\n"); //$NON-NLS-1$ //$NON-NLS-2$
		info.append(Messages.getString("ConnectionServerDriver_61") + "\r\n\r\n"); //$NON-NLS-1$ //$NON-NLS-2$
		for (idx = 0; idx < objects.size(); idx++) {
			info.append(Messages.getString("ConnectionServerDriver_63") + objects.get(idx).getFieldValue("CODE") + "\r\n" + //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
									Messages.getString("ConnectionServerDriver_64") + objects.get(idx).getFieldValue("STATUS") + "\r\n" + //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
									Messages.getString("ConnectionServerDriver_65") + objects.get(idx).getFieldValue("DSCR") + "\r\n" + //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
									Messages.getString("ConnectionServerDriver_66") + objects.get(idx).getFieldValue("VVRM") + " (" + objects.get(idx).getFieldValue("PID") + ")" + "\r\n" + //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$ //$NON-NLS-6$
									Messages.getString("ConnectionServerDriver_67") + objects.get(idx).getFieldValue("APAR") + "/" + objects.get(idx).getFieldValue("MODSL") + "\r\n"); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$
			if (idx < (objects.size() - 1)) {
				info.append("\r\n\r\n"); //$NON-NLS-1$
			}
		}
		
		return info.toString();
	}
	
	public String runCommand(String cmd, String productId) {
		
		return runCommand(cmd, productId, null, null);
	}
	
	public String runCommand(String cmd, String productId, Xheader previousXhdr, String cFunc) {
		
		ConnectionProfile profile = null;
		boolean needToRunCommand = true;
		String response = "";
		Socket socket = null;
		int idx;
		
		/*
		 *  If the connection profile is marked not useable, report it, otherwise run the command.
		 *  
		 *  If the command failed due to a userid/password related issue, try to get new credentials and rerun it.
		 */
		
		/*
		 *  Get the profile each time, in case it has been altered by the user since last use.
		 *  
		 *  The server may not actually exist in a persisted state (e.g. user has changed in the dialog but not clicked Apply or Save.  If it doesn't exist,
		 *  use the supplied host/port.  In this case the connection profile must already exist, as they are maintained separately.
		 */
		if (instance.getConnectionServer(srvr.getName(), srvr.getHost(), srvr.getPort()) == null) {
			profile = instance.getConnectionProfile(srvr.getConnectionProfileName());
		}
		else {
			profile = instance.getConnectionProfile(instance.getConnectionServer(srvr.getName(), srvr.getHost(), srvr.getPort()).getConnectionProfileName());
		}
		
		if (!profile.isUseable()) {
			instance.addConsoleMessage(new ConsoleMessage(srvr.getFullTitle() + ":  " + Messages.getString("ConnectionServerDriver_56") + profile.getName() + Messages.getString("ConnectionServerDriver_57") + profile.getUnuseableReason())); //$NON-NLS-1$ //$NON-NLS-2$
			srvr.setConnectStatus(AppInstance.CAN_CONNECT_AUTHENTICATION_PROBLEM);
			return "";
		}
		else {
		
			needToRunCommand = true;
			while (needToRunCommand) {
				socket = openConnection(previousXhdr);
				//Set timeout for a PING, as just detecting the server is responding, it should be quick, and the socket won't be reused.
				if (cmd.toLowerCase().equals("ping")) {   //$NON-NLS-1$ 
					try {
						socket.setSoTimeout(20000);  //20 seconds.
					}
					catch (SocketException e) {}  //Don't care much.
				}
				if (socket != null) {
					send(socket, cmd, productId, previousXhdr, cFunc);
					receiveFields(socket);
					if (cFunc == null) {  //Can't be in conversational mode.
						closeConnection(socket);
					}
					else {
						if (cFunc.equals("4")) {
							synchronized (this) {
								sockets.add(socket);
								socketIds.add(new String(xhdr.getXhdrSConvTkn()));
							}
						}
						else {
							if (cFunc.equals("5")) {
								synchronized (this) {
									idx = socketIds.indexOf(new String(previousXhdr.getXhdrSConvTkn()));
									if (idx > 1) {
										socketIds.remove(idx);
										sockets.remove(idx);
									}
								}
							}
						}
					}
				}
				response = getAppResponse();
				if (!profile.isUseable()) {
					srvr.setConnectStatus(AppInstance.CAN_CONNECT_AUTHENTICATION_PROBLEM);
					instance.notifyConnectionProfileEditListeners(profile);  //This can reset profile.isUseable().
					if (!profile.isUseable()) {
						needToRunCommand = false;  //Profile edit listener didn't reset the credentials.
					}
				}
				else {
					needToRunCommand = false;
				}
			}
		}

		return appResponse;
	}
	
	private Socket openConnection(Xheader xhdr) {
		
		return (Socket) openConnection(xhdr, false);
	}
	
	private Object openConnection(Xheader xhdr, boolean forIBMExplorer) {

		Socket socket = null;
		int idx;
		
		//If in conversational mode (got a prior xhdr to continue with), just return the socket for that conversation,
		//otherwise return a new socket.
		if (xhdr != null && !forIBMExplorer) {
			idx = socketIds.indexOf(new String(xhdr.getXhdrSConvTkn()));
			if (idx > -1) {
				socket = sockets.get(idx);
				return socket;
			}
		}
		
		try {
			socket = new Socket(srvr.getHost(), srvr.getPort());
			socket.setTcpNoDelay(true);
			socket.setKeepAlive(true);
		}
		catch (Exception e1) {
			//If opening for IBM Explorer, we want to return the exception so that it can be handled up the chain.
			srvr.setConnectStatus(AppInstance.CAN_CONNECT_UNABLE);
			if (forIBMExplorer) {
				return e1;
			}
			else {
				instance.addConsoleMessage(new ConsoleMessage(Messages.getString("ConnectionServerDriver_6") + srvr.getFullTitle())); //$NON-NLS-1$
				instance.addConsoleMessage(e1);
				return null;
			}
		}
		
		return socket;
	}
	
	public int testConnection() {
	
		return (Integer) testConnection(false);
	}
	
	public Object testConnection(boolean forIBMExplorer) {
		
		Object ret;  //Can be a socket, or an Exception if testing for IBM Explorer.
		
		srvr.setConnectStatus(AppInstance.CAN_CONNECT_OK);
		ret = openConnection(null, forIBMExplorer);
		if (srvr.getConnectStatus() == AppInstance.CAN_CONNECT_OK) {
			runCommand("PING", AppInstance.fudProductId); //$NON-NLS-1$
			closeConnection((Socket) ret);
		}
		else {
			if (forIBMExplorer && Exception.class.isInstance(ret)) {
				//srvr.setConnectFailureReason(((Exception) ret).getMessage());
				return ret;
			}
		}
		
		if (srvr.getConnectStatus() == AppInstance.CAN_CONNECT_OK) {
			if (xhdrRc == 0 && xhdrRsn == 0) {
				return AppInstance.CAN_CONNECT_OK;
			}
			else {
				return AppInstance.CAN_CONNECT_UNABLE;
			}
		}
		else {
			return srvr.getConnectStatus();
		}
	}
	
	private void closeConnection(Socket socket) {
		
		try {
			socket.close();
		}
		catch (IOException e) {
			instance.addConsoleMessage(new ConsoleMessage(Messages.getString("ConnectionServerDriver_7") + srvr.getFullTitle())); //$NON-NLS-1$
			instance.addConsoleMessage(e);
		}
	}
	
	private String send(Socket socket, String cmd, String product, Xheader previousXhdr, String cFunc) {
		
		ByteArrayOutputStream messageBuffer;
		DataOutputStream buf;
		DataOutputStream out = null;
		
		initialiseXhdr(previousXhdr, cFunc);
		
		xhdrTotalMsgLength = 260 + cmd.length();  //260 is the header length including these 4 length bytes.
		
		xhdrCProduct = product;
		
		//Encrypt the userid and password. //FUD-590: CSL Server now uppercases userid, may as well match this behaviour here. CEX already uppercases userid
		xhdrCUid = encrypt(new StringEditor(instance.getConnectionUserid(srvr.getConnectionProfileName()).toUpperCase()).pad(8), xhdrTotalMsgLength);
		xhdrCPwd = encrypt(new StringEditor(instance.getConnectionPassword(srvr.getConnectionProfileName())).pad(8), xhdrTotalMsgLength);
		
		try {
			// Allocate a conveniently sized byte stream. 	
			messageBuffer = new ByteArrayOutputStream(xhdrTotalMsgLength);
			// Associate a DataOutputStream to this newly created byte stream.
			buf = new DataOutputStream(messageBuffer);
		
			out = new DataOutputStream(socket.getOutputStream());
		
			buf.writeInt(xhdrTotalMsgLength);
			buf.writeBytes(xhdrCtlFldEyec);
			buf.writeBytes(xhdrCtlFldArclev);
			buf.writeBytes(xhdrCtlFldType);
			buf.writeInt(xhdrTcpTMTE);
			buf.write(xhdrTcpSKD);
			buf.writeShort(xhdrTcpPort);
			buf.writeBytes(xhdrTcpAddr);
			buf.writeInt(xhdrSdLen);
			buf.writeBytes(xhdrSdSegmctl);
			buf.writeBytes(xhdrSdTr);
			buf.writeBytes(xhdrCFunc);
			buf.writeBytes(xhdrCProduct);
			buf.writeShort(xhdrCCCSID);
			buf.writeInt(xhdrCBufsz);
			buf.writeBytes(xhdrCChkid);
			buf.writeBytes(xhdrCUid);
			buf.writeBytes(xhdrCPwd);
			buf.writeLong(xhdrCReserved1);
			buf.write(xhdrSConvTkn);
			buf.write(xhdrRRSESTkn);
			buf.write(xhdrSPRDE);
			buf.write(xhdrSConv);
			buf.write(xhdrSRSES);
			buf.writeInt(xhdrRc);
			buf.writeInt(xhdrRsn);
			buf.writeInt(xhdrInfo);
			buf.writeInt(xhdrInfo2);
			buf.writeInt(xhdrReserved2);
			buf.writeBytes(xhdrCRc);
			buf.writeBytes(xhdrCRsn);
			buf.writeBytes(xhdrCInfo);
			buf.writeBytes(xhdrCInfo1);
			buf.writeLong(xhdrCReserved3);
			buf.writeBytes(xhdrStaTime);
			buf.writeBytes(xhdrStaLSN);
			buf.writeBytes(xhdrStoTime);
			buf.writeBytes(xhdrStoLSN);
		
			buf.writeBytes(cmd);
		
			buf.flush();
		
			//System.out.println(new String(messageBuffer.toByteArray()));
			out.write(messageBuffer.toByteArray());
			out.flush();
		}
		catch (Throwable e) {
			instance.addConsoleMessage(new ConsoleMessage(Messages.getString("ConnectionServerDriver_8") + srvr.getFullTitle())); //$NON-NLS-1$
			instance.addConsoleMessage(e);
			return instance.exceptionToString(e);
		}
		
		return ""; //$NON-NLS-1$
	}
	
	private void initialiseXhdr(Xheader previousXhdr, String cFunc) {
		
		xhdrTotalMsgLength = 0;  //Total length of message, including payload.
		if (previousXhdr == null) {
			//XHDR control fields.
			xhdrCtlFldEyec = Messages.getString("ConnectionServerDriver_10");  //Always use XHDR (if you use ASCII, it will perform translation of XHDR) //$NON-NLS-1$
			xhdrCtlFldArclev = Messages.getString("ConnectionServerDriver_11"); //Architecture level: use 0 //$NON-NLS-1$
			xhdrCtlFldType = Messages.getString("ConnectionServerDriver_12"); //Always use TCP //$NON-NLS-1$
			//Fields set by server for TCP/IP client (do not set or clear, just skip 24 Bytes)
			xhdrTcpTMTE = 0;  //TCP Message Thread Element
			xhdrTcpSKD = new byte[2];  //Local socket descriptor
			xhdrTcpPort = 0; //Remote port number
			xhdrTcpAddr = new StringEditor(Messages.getString("ConnectionServerDriver_13")).pad(16);  //Remote INET address //$NON-NLS-1$
			//Session data the client needs to set
			xhdrSdLen = 0;  //Estimated length of message (optional, helps the server allocate memory)
			xhdrSdSegmctl = Messages.getString("ConnectionServerDriver_14");  //F : First;  M : Middle  (not first or last); L : Last //$NON-NLS-1$
			xhdrSdTr = Messages.getString("ConnectionServerDriver_15");  //Y : Yes, payload requires translation (Y if using ASCII); or N for no translation //$NON-NLS-1$
			//Fields the client needs to set
			if (cFunc == null) {
				xhdrCFunc = Messages.getString("ConnectionServerDriver_16");  //Always use 1 - Product request //$NON-NLS-1$
			}
			else {
				xhdrCFunc = cFunc;
			}
			xhdrCProduct = new StringEditor(Messages.getString("ConnectionServerDriver_17")).pad(3);  //E.g. FUD or GPL (for ping and delay) //$NON-NLS-1$
			xhdrCCCSID = 819; //819 = x'333'.  //55;  //Use x'37'
			xhdrCBufsz = 0;  //Buffer size use 0 to get each response in full, i.e. unsegmented.
			xhdrCChkid = new StringEditor(Messages.getString("ConnectionServerDriver_18")).pad(16);  //Can set to zeros (it is echoed back so you can use it for consistency checking) //$NON-NLS-1$
			xhdrCUid = new StringEditor(Messages.getString("ConnectionServerDriver_19")).pad(8);  //User ID (use the current CEX encryption for the user ID) //$NON-NLS-1$
			xhdrCPwd = new StringEditor(Messages.getString("ConnectionServerDriver_20")).pad(8);  //Password (use the current CEX encryption for the password) //$NON-NLS-1$
			xhdrCReserved1 = 0;  //For future use
			//Server-maintained fields client must keep and send back on requests (28 bytes)
			xhdrSConvTkn = new byte[8];  //CONV Token
			xhdrRRSESTkn = new byte[8];  //RSES Token
			xhdrSPRDE = new byte[4];  //PRDE entry address
			xhdrSConv = new byte[4];  //CONV address
			xhdrSRSES = new byte[4];  //RSES entry address
			//Server status information (view only, for diag use information in next section, not this)
			xhdrRc = 0;
			xhdrRsn = 0;
			xhdrInfo = 0;
			xhdrInfo2 = 0;
			xhdrReserved2 = 0;
			//Server feedback in text (40 bytes)
			xhdrCRc = new StringEditor(Messages.getString("ConnectionServerDriver_21")).pad(8);  //Return code //$NON-NLS-1$
			xhdrCRsn = new StringEditor(Messages.getString("ConnectionServerDriver_22")).pad(8);  //Word 1 = <ModuleID><ReasonCode> //$NON-NLS-1$
			xhdrCInfo = new StringEditor(Messages.getString("ConnectionServerDriver_23")).pad(8); //INFO //$NON-NLS-1$
			xhdrCInfo1 = new StringEditor(Messages.getString("ConnectionServerDriver_24")).pad(8);  //INFO //$NON-NLS-1$
			xhdrCReserved3 = 0;
			//Server request-response cycle statistics fields
			xhdrStaTime = new StringEditor(Messages.getString("ConnectionServerDriver_25")).pad(24);  //Timestamp of XHDR read into server //$NON-NLS-1$
			xhdrStaLSN = new StringEditor(Messages.getString("ConnectionServerDriver_26")).pad(16);  //with server log sequence number //$NON-NLS-1$
			xhdrStoTime = new StringEditor(Messages.getString("ConnectionServerDriver_27")).pad(24);  //Timestamp of XHDR send from server //$NON-NLS-1$
			xhdrStoLSN = new StringEditor(Messages.getString("ConnectionServerDriver_28")).pad(16);  //with server log sequence number //$NON-NLS-1$
		}
		else {
			//XHDR control fields.
			xhdrCtlFldEyec = previousXhdr.getXhdrCtlFldEyec();
			xhdrCtlFldArclev = previousXhdr.getXhdrCtlFldArclev();
			xhdrCtlFldType = previousXhdr.getXhdrCtlFldType();
			xhdrTcpTMTE = previousXhdr.getXhdrTcpTMTE();
			xhdrTcpSKD = previousXhdr.getXhdrTcpSKD();
			xhdrTcpPort = previousXhdr.getXhdrTcpPort();
			xhdrTcpAddr = previousXhdr.getXhdrTcpAddr();
			xhdrSdLen = 0;  //Estimated length of message (optional, helps the server allocate memory)
			xhdrSdSegmctl = previousXhdr.getXhdrSdSegmctl();
			xhdrSdTr = previousXhdr.getXhdrSdTr();
			xhdrCFunc = cFunc;
			xhdrCProduct = previousXhdr.getXhdrCProduct();
			xhdrCCCSID = previousXhdr.getXhdrCCCSID();
			xhdrCBufsz = 0;  //Buffer size use 0 to get each response in full, i.e. unsegmented.
			xhdrCChkid = previousXhdr.getXhdrCChkid();
			xhdrCUid = previousXhdr.getXhdrCUid();
			xhdrCPwd = new StringEditor(Messages.getString("ConnectionServerDriver_20")).pad(8);  //Password (use the current CEX encryption for the password) //$NON-NLS-1$
			xhdrCReserved1 = 0;  //For future use
			//Server-maintained fields client must keep and send back on requests (28 bytes)
			xhdrSConvTkn = previousXhdr.getXhdrSConvTkn();
			xhdrRRSESTkn = previousXhdr.getXhdrRRSESTkn();
			xhdrSPRDE = previousXhdr.getXhdrSPRDE();
			xhdrSConv = previousXhdr.getXhdrSConv();
			xhdrSRSES = previousXhdr.getXhdrSRSES();
			//Server status information (view only, for diag use information in next section, not this)
			xhdrRc = 0;
			xhdrRsn = 0;
			xhdrInfo = 0;
			xhdrInfo2 = 0;
			xhdrReserved2 = 0;
			//Server feedback in text (40 bytes)
			xhdrCRc = new StringEditor(Messages.getString("ConnectionServerDriver_21")).pad(8);  //Return code //$NON-NLS-1$
			xhdrCRsn = new StringEditor(Messages.getString("ConnectionServerDriver_22")).pad(8);  //Word 1 = <ModuleID><ReasonCode> //$NON-NLS-1$
			xhdrCInfo = new StringEditor(Messages.getString("ConnectionServerDriver_23")).pad(8); //INFO //$NON-NLS-1$
			xhdrCInfo1 = new StringEditor(Messages.getString("ConnectionServerDriver_24")).pad(8);  //INFO //$NON-NLS-1$
			xhdrCReserved3 = 0;
			//Server request-response cycle statistics fields
			xhdrStaTime = new StringEditor(Messages.getString("ConnectionServerDriver_25")).pad(24);  //Timestamp of XHDR read into server //$NON-NLS-1$
			xhdrStaLSN = new StringEditor(Messages.getString("ConnectionServerDriver_26")).pad(16);  //with server log sequence number //$NON-NLS-1$
			xhdrStoTime = new StringEditor(Messages.getString("ConnectionServerDriver_27")).pad(24);  //Timestamp of XHDR send from server //$NON-NLS-1$
			xhdrStoLSN = new StringEditor(Messages.getString("ConnectionServerDriver_28")).pad(16);  //with server log sequence number //$NON-NLS-1$
		}
	}
	
	private void receiveFields(Socket socket) {
		
		DataInputStream in = null;
		byte[] msgBytes = null;
		byte[] bytes1 = new byte[1];
		byte[] bytes3 = new byte[3];
		byte[] bytes4 = new byte[4];
		byte[] bytes8 = new byte[8];
		byte[] bytes16 = new byte[16];
		byte[] bytes24 = new byte[24];
		String srvrMsgType = ""; //$NON-NLS-1$
		String errorMsg = ""; //$NON-NLS-1$
		
		appResponse = null;
		
		if (socket == null) {
			return;
		}
		
		try {
			in = new DataInputStream(socket.getInputStream());
			
			xhdrTotalMsgLength = in.readInt();
			
			//Read a bit of the message to see if it's probably a format we recognise.
			//If it isn't, don't get any more, in case the length is corrupted and too big to read into memory.
			if (xhdrTotalMsgLength >= xhdrLength) {  //The XHDR should be 256 bytes.
				xhdr = new Xheader();
				
				in.read(bytes4, 0, 4);
				xhdrCtlFldEyec = new String(bytes4);
				if (!(xhdrCtlFldEyec.equals(Messages.getString("ConnectionServerDriver_10")))) {
					xhdrRc = 12;  //Set it, because if not to a valid Connection Server port it won't be.
					instance.addConsoleMessage(new ConsoleMessage(srvrMsgType + Messages.getString("ConnectionServerDriver_44") + srvr.getFullTitle() + //$NON-NLS-1$
																								Messages.getString("ConnectionServerDriver_45"))); //$NON-NLS-1$
					srvr.setConnectFailureReason(srvrMsgType + Messages.getString("ConnectionServerDriver_44") + srvr.getFullTitle() + //$NON-NLS-1$
							Messages.getString("ConnectionServerDriver_45"));
					return;
				}
				in.read(bytes1, 0, 1);
				xhdrCtlFldArclev = new String(bytes1);
				in.read(bytes3, 0, 3);
				xhdrCtlFldType = new String(bytes3);
				xhdrTcpTMTE = in.readInt();
				//xhdrTcpSKD = in.readShort();
				in.read(xhdrTcpSKD,0, 2);
				xhdrTcpPort = in.readShort();
				in.read(bytes16, 0, 16);
				xhdrTcpAddr = new String(bytes16);
				xhdrSdLen = in.readInt();
				in.read(bytes1, 0, 1);
				xhdrSdSegmctl = new String(bytes1);
				in.read(bytes1, 0, 1);
				xhdrSdTr = new String(bytes1);
				in.read(bytes1, 0, 1);
				xhdrCFunc = new String(bytes1);
				in.read(bytes3, 0, 3);
				xhdrCProduct = new String(bytes3);
				xhdrCCCSID = in.readShort();
				xhdrCBufsz = in.readInt();
				in.read(bytes16, 0, 16);
				xhdrCChkid = new String(bytes16);
				in.read(bytes8, 0, 8);
				xhdrCUid = new String(bytes8);
				in.read(bytes8, 0, 8);
				xhdrCPwd = new String(bytes8);
				xhdrCReserved1 = in.readLong();
				//xhdrSConvTkn = in.readLong();
				in.read(xhdrSConvTkn,0, 8);
				//xhdrRRSESTkn = in.readLong();
				in.read(xhdrRRSESTkn,0, 8);
				//xhdrSPRDE = in.readInt();
				in.read(xhdrSPRDE,0, 4);
				//xhdrSConv = in.readInt();
				in.read(xhdrSConv,0, 4);
				//xhdrSRSES = in.readInt();
				in.read(xhdrSRSES,0, 4);
				xhdrRc = in.readInt();
				xhdrRsn = in.readInt();
				xhdrInfo = in.readInt();
				xhdrInfo2 = in.readInt();
				xhdrReserved2 = in.readInt();
				in.read(bytes8, 0, 8);
				xhdrCRc = new String(bytes8);
				in.read(bytes8, 0, 8);
				xhdrCRsn = new String(bytes8);
				in.read(bytes8, 0, 8);
				xhdrCInfo = new String(bytes8);
				in.read(bytes8, 0, 8);
				xhdrCInfo1 = new String(bytes8);
				xhdrCReserved3 = in.readLong();
				in.read(bytes24, 0, 24);
				xhdrStaTime = new String(bytes24);
				in.read(bytes16, 0, 16);
				xhdrStaLSN = new String(bytes16);
				in.read(bytes24, 0, 24);
				xhdrStoTime = new String(bytes24);
				in.read(bytes16, 0, 16);
				xhdrStoLSN = new String(bytes16);
			
				if ((xhdrTotalMsgLength - xhdrLength - 4) > 0) {
					msgBytes = new byte[xhdrTotalMsgLength - xhdrLength - 4];
					in.readFully(msgBytes);
					appResponse = new String(msgBytes);
					srvr.setConnectFailureReason(appResponse);
				}
			}
			else {
				msgBytes = new byte[xhdrTotalMsgLength - 4];
				in.readFully(msgBytes);
				appResponse = new String(msgBytes);
				srvr.setConnectFailureReason(appResponse);
			}
			
			//Ignore the response if the server didn't support the product - shouldn't happen but we may not be able to detect it.
			if (xhdrRc  > 0 && !(appResponse.startsWith(Messages.getString("ConnectionServerDriver_46")))) {
				if (xhdrRc == 4) {
					srvrMsgType = Messages.getString("ConnectionServerDriver_30"); //$NON-NLS-1$
				}
				else {
					srvrMsgType = Messages.getString("ConnectionServerDriver_31"); //$NON-NLS-1$
				}
				instance.addConsoleMessage(new ConsoleMessage(srvrMsgType + Messages.getString("ConnectionServerDriver_32") + srvr.getFullTitle())); //$NON-NLS-1$
				errorMsg = Messages.getString("ConnectionServerDriver_33") + xhdrCRc + Messages.getString("ConnectionServerDriver_34") + xhdrCRsn + Messages.getString("ConnectionServerDriver_35") + xhdrCInfo + Messages.getString("ConnectionServerDriver_36") + xhdrCInfo1; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$
				srvr.setConnectFailureReason(Messages.getString("ConnectionServerDriver_33") + xhdrCRc + Messages.getString("ConnectionServerDriver_34") + xhdrCRsn + Messages.getString("ConnectionServerDriver_35") + xhdrCInfo + Messages.getString("ConnectionServerDriver_36") + xhdrCInfo1); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$
				if (xhdrRsn > 0) {
					errorMsg = errorMsg + Messages.getString("ConnectionServerDriver_43") + appResponse; //$NON-NLS-1$
					srvr.setConnectFailureReason(appResponse);
					
					if (appResponse.startsWith(Messages.getString("ConnectionServerDriver_51")) || //$NON-NLS-1$
							appResponse.startsWith(Messages.getString("ConnectionServerDriver_52")) || //$NON-NLS-1$
							appResponse.startsWith(Messages.getString("ConnectionServerDriver_53")) || //$NON-NLS-1$
							appResponse.startsWith(Messages.getString("ConnectionServerDriver_54")) || //$NON-NLS-1$
							appResponse.startsWith(Messages.getString("ConnectionServerDriver_55")) || //$NON-NLS-1$
							appResponse.startsWith(Messages.getString("ConnectionServerDriver_58"))) { //$NON-NLS-1$
						instance.getConnectionProfile(srvr.getConnectionProfileName()).setUnuseable(appResponse.trim());
						srvr.setConnectStatus(AppInstance.CAN_CONNECT_AUTHENTICATION_PROBLEM);
					}
					
					appResponse = null;
				}
				instance.addConsoleMessage(new ConsoleMessage(errorMsg));
			}
			
			//in.close();
			in = null;
		}
		catch (IOException e) {
			instance.addConsoleMessage(new ConsoleMessage(Messages.getString("ConnectionServerDriver_37") + srvr.getFullTitle())); //$NON-NLS-1$
			instance.addConsoleMessage(e);
			xhdrRc = 12;  //Set it, because if not to a valid Connection Server port it won't be.
			srvr.setConnectFailureReason(e.getMessage());
			return;
		}
		
		xhdr.setXhdrCtlFldEyec(xhdrCtlFldEyec);
		xhdr.setXhdrCtlFldArclev(xhdrCtlFldArclev);
		xhdr.setXhdrCtlFldType(xhdrCtlFldType);
		xhdr.setXhdrTcpTMTE(xhdrTcpTMTE);
		xhdr.setXhdrTcpSKD(xhdrTcpSKD);
		xhdr.setXhdrTcpPort(xhdrTcpPort);
		xhdr.setXhdrTcpAddr(xhdrTcpAddr);
		xhdr.setXhdrSdLen(xhdrSdLen);
		xhdr.setXhdrSdSegmctl(xhdrSdSegmctl);
		xhdr.setXhdrSdTr(xhdrSdTr);
		xhdr.setXhdrCFunc(xhdrCFunc);
		xhdr.setXhdrCProduct(xhdrCProduct);
		xhdr.setXhdrCCCSID(xhdrCCCSID);
		xhdr.setXhdrCBufsz(xhdrCBufsz);
		xhdr.setXhdrCChkid(xhdrCChkid);
		xhdr.setXhdrCUid(xhdrCUid);
		xhdr.setXhdrCPwd(xhdrCPwd);
		xhdr.setXhdrCReserved1(xhdrCReserved1);
		xhdr.setXhdrSConvTkn(xhdrSConvTkn);
		xhdr.setXhdrRRSESTkn(xhdrRRSESTkn);
		xhdr.setXhdrSPRDE(xhdrSPRDE);
		xhdr.setXhdrSConv(xhdrSConv);
		xhdr.setXhdrSRSES(xhdrSRSES);
		xhdr.setXhdrRc(xhdrRc);
		xhdr.setXhdrRsn(xhdrRsn);
		xhdr.setXhdrInfo(xhdrInfo);
		xhdr.setXhdrInfo2(xhdrInfo2);
		xhdr.setXhdrReserved2(xhdrReserved2);
		xhdr.setXhdrCRc(xhdrCRc);
		xhdr.setXhdrCRsn(xhdrCRsn);
		xhdr.setXhdrCInfo(xhdrCInfo);
		xhdr.setXhdrCInfo1(xhdrCInfo1);
		xhdr.setXhdrCReserved3(xhdrCReserved3);
		xhdr.setXhdrStaTime(xhdrStaTime);
		xhdr.setXhdrStaLSN(xhdrStaLSN);
		xhdr.setXhdrStoTime(xhdrStoTime);
		xhdr.setXhdrStoLSN(xhdrStoLSN);
	}
	
	private void receive(Socket socket) {
		
		DataInputStream in = null;
		int msgLen;
		byte[] msgBytes = null;
		byte[] xhdrBytes = new byte[xhdrLength];
		int cnt;
		String xhdr = null;
		
		appResponse = null;
		
		if (socket == null) {
			return;
		}
		
		try {
			in = new DataInputStream(socket.getInputStream());
			
			msgLen = in.readInt() - 4;
			
			//Read a bit of the message to see if it's probably a format we recognise.
			//If it isn't, don't get any more, in case the length is corrupted and too big to read into memory.
			if (msgLen >= xhdrLength) {  //The XHDR should be 256 bytes.
				in.read(xhdrBytes, 0, xhdrLength);
				xhdr = new String(xhdrBytes);
				
				if (xhdr.substring(0, 4).equals(Messages.getString("ConnectionServerDriver_38")) && xhdr.substring(4, 7).equals(Messages.getString("ConnectionServerDriver_39"))) { //$NON-NLS-1$ //$NON-NLS-2$
					//Read the rest.
					msgBytes = new byte[msgLen];
					in.readFully(msgBytes, xhdrLength, msgLen - xhdrLength);
				}
				else {
					msgBytes = new byte[xhdrLength];
				}
				//Put the prefix already read back in.
				for (cnt = 0; cnt < xhdrLength; cnt++) {
					msgBytes[cnt] = xhdrBytes[cnt];
				}
			}
			else {
				msgBytes = new byte[msgLen];
				in.readFully(msgBytes);
			}
		
			appResponse = new String(msgBytes);
			//System.out.println(appResponse);
			
			in.close();
		}
		catch (IOException e) {
			instance.addConsoleMessage(new ConsoleMessage(Messages.getString("ConnectionServerDriver_40") + srvr.getFullTitle())); //$NON-NLS-1$
			instance.addConsoleMessage(e);
		}
	}
	
	private String getAppResponse() {
		
		return appResponse;
	}
	
	public Xheader getXhdr() {
		
		return xhdr;
	}

	public static String encrypt(String origValue, int cycleLength) {
		//refer FUD000.QAAUTO.$QAAEXEC(FUDCLNT3) 
		if (origValue == null) {	//assembler code returns low values for low values so returning null for null here
			return null;
		}

		//pre-conditions
		if (cycleLength <= 0) throw new IllegalArgumentException("cycleLength");									//assembler code requires positive cycle length
		if (origValue.length() == 0 || origValue.length() > 255) throw new IllegalArgumentException("origValue");	//assembler code requires length in 1..256

		//obfuscate in multiple parses (character sets must be disjoint)
		final String[] allCharSets = new String[]{
				"wBP9JyM0QIX3Lh@2AKCeW t16koNvbRzGrU$spi7V4lfxEOSdFZguqHn5j#cD8TmYa",	//$NON-NLS-1$
				"+|-%=:<?_&!.>*"														//$NON-NLS-1$
		}; 						
		char[] inputChars = origValue.toCharArray();
		for (String charSet: allCharSets) {
			int offset = cycleLength % charSet.length();
			
			for (int i = 0; i < inputChars.length; i++) {
				char origChar = inputChars[i];
				int origPos = charSet.indexOf(origChar);
				if (origPos != -1) {
					inputChars[i] =	charSet.charAt(
										(origPos + offset) % charSet.length()
									);
					offset += 1;
				}
			}
		}
		
		return String.valueOf(inputChars);
	}
}
